package net.sf.doolin.context.spring;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import net.sf.doolin.util.Pair;
import net.sf.doolin.util.xml.DOMUtils;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.Validate;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.beans.factory.config.BeanDefinitionHolder;
import org.springframework.beans.factory.support.BeanDefinitionBuilder;
import org.springframework.beans.factory.xml.BeanDefinitionDecorator;
import org.springframework.beans.factory.xml.ParserContext;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * Customisable parser for custom elements. It can be used both as
 * {@linkplain org.springframework.beans.factory.xml.BeanDefinitionParser
 * parser} or as a {@linkplain BeanDefinitionDecorator decorator}.
 * 
 * @author Damien Coraboeuf
 * 
 * @param <C>
 *            Type of bean which is created by this parser.
 */
public class SimpleDefinitionParser<C> extends AbstractDefinitionParser
		implements BeanDefinitionDecorator {

	/**
	 * Creates an instance of a parser for the given bean class
	 * 
	 * @param <T>
	 *            Type of parser to create
	 * @param beanClass
	 *            Class of the bean to create
	 * @return Parser
	 */
	public static <T> SimpleDefinitionParser<T> getInstance(Class<T> beanClass) {
		return new SimpleDefinitionParser<T>(beanClass);
	}

	private final Class<C> beanClass;

	private String defaultID;
	private Map<String, String> customBeanAs = new HashMap<String, String>();
	private Map<String, String> simpleAttributes = new HashMap<String, String>();
	private Map<String, String> fixedProperties = new HashMap<String, String>();
	private Map<String, String> attributeRefs = new HashMap<String, String>();

	private Map<String, List<Pair<String, SimpleBeanFactory>>> attributeFactories = new HashMap<String, List<Pair<String, SimpleBeanFactory>>>();
	private Pair<String, String> listContainer;
	private Pair<String, String> listStringContainer;
	private Pair<String, String> mapContainer;

	private String containerProperty;

	/**
	 * Constructor
	 * 
	 * @param beanClass
	 *            Class of the object to create
	 */
	public SimpleDefinitionParser(Class<C> beanClass) {
		this.beanClass = beanClass;
	}

	/**
	 * Uses the value of an attribute as a reference to a bean. The bean is
	 * stored in a property that has the same name than the attribute.
	 * 
	 * @param attributeName
	 *            Attribute name
	 * @return This parser (allows chaining)
	 */
	public SimpleDefinitionParser<C> addAttributeRef(String attributeName) {
		return addAttributeRef(attributeName, attributeName);
	}

	/**
	 * Uses the value of an attribute as a reference to a bean. The bean is
	 * stored in a property.
	 * 
	 * @param attributeName
	 *            Attribute name
	 * @param propertyName
	 *            Property name
	 * @return This parser (allows chaining)
	 */
	public SimpleDefinitionParser<C> addAttributeRef(String attributeName,
			String propertyName) {
		this.attributeRefs.put(attributeName, propertyName);
		return this;
	}

	/**
	 * Registers a child element which must be parsed as a custom element.
	 * 
	 * @param customBeanName
	 *            Name of the custom element.
	 * @param propertyName
	 *            Property that must hold the parsed custom element.
	 * @return This parser (allows chaining)
	 */
	public SimpleDefinitionParser<C> addCustomBeanAs(String customBeanName,
			String propertyName) {
		this.customBeanAs.put(customBeanName, propertyName);
		return this;
	}

	/**
	 * Specifies that the value of an attribute must be used to create an object
	 * that will be used as a property value.
	 * 
	 * @param attributeName
	 *            Attribute name
	 * @param propertyName
	 *            Property name
	 * @param beanFactory
	 *            Factory that creates an object from the value of the attribute
	 * @return This parser (allows chaining)
	 */
	public SimpleDefinitionParser<C> addFactoryFromAttribute(
			String attributeName, String propertyName,
			SimpleBeanFactory beanFactory) {
		List<Pair<String, SimpleBeanFactory>> list = this.attributeFactories
				.get(attributeName);
		if (list == null) {
			list = new ArrayList<Pair<String, SimpleBeanFactory>>();
			this.attributeFactories.put(attributeName, list);
		}
		list
				.add(new Pair<String, SimpleBeanFactory>(propertyName,
						beanFactory));
		return this;
	}

	/**
	 * Specifies the value for a property
	 * 
	 * @param propertyName
	 *            Property name
	 * @param propertyValue
	 *            Property value
	 * @return This parsing (allows chaining)
	 */
	public SimpleDefinitionParser<C> addPropertyValue(String propertyName,
			String propertyValue) {
		this.fixedProperties.put(propertyName, propertyValue);
		return this;
	}

	/**
	 * Specifies that the value of the given attribute must be used as the value
	 * for the property that has the same name
	 * 
	 * @param attributeName
	 *            Attribute name
	 * @return This parser (allows chaining)
	 */
	public SimpleDefinitionParser<C> addSimpleAttribute(String attributeName) {
		return addSimpleAttribute(attributeName, attributeName);
	}

	/**
	 * Specifies that the value of the given attribute must be used as the value
	 * of a property.
	 * 
	 * @param attributeName
	 *            Attribute name
	 * @param propertyName
	 *            Property name
	 * @return This parser (allows chaining)
	 */
	public SimpleDefinitionParser<C> addSimpleAttribute(String attributeName,
			String propertyName) {
		this.simpleAttributes.put(attributeName, propertyName);
		return this;
	}

	@Override
	public BeanDefinitionHolder decorate(Node node,
			BeanDefinitionHolder holder, ParserContext parserContext) {
		// Checks the containerProperty
		Validate
				.notEmpty(
						this.containerProperty,
						String
								.format(
										"containerProperty must be defined for the parser of %s when used as a decorator",
										this.beanClass.getSimpleName()));
		// Element
		Element element = (Element) node;
		// Parses this element
		BeanDefinition definition;
		String oldContainerProperty = this.containerProperty;
		try {
			this.containerProperty = null;
			ParserContext customParserContext = new ParserContext(parserContext
					.getReaderContext(), parserContext.getDelegate(), holder
					.getBeanDefinition());
			definition = parse(element, customParserContext);
		} finally {
			this.containerProperty = oldContainerProperty;
		}
		// Sets the property
		holder.getBeanDefinition().getPropertyValues().addPropertyValue(
				this.containerProperty, definition);
		// OK
		return holder;
	}

	@Override
	public void doParse(Element element, ParserContext parserContext,
			BeanDefinitionBuilder builder) {
		// Parsing of sub-elements (does not parse custom elements)
		parserContext.getDelegate().parsePropertyElements(element,
				builder.getBeanDefinition());
		// Custom beans
		for (Map.Entry<String, String> entry : this.customBeanAs.entrySet()) {
			String customBeanName = entry.getKey();
			String propertyName = entry.getValue();
			Element eCustomBean = DOMUtils.getElement(element, customBeanName);
			if (eCustomBean != null) {
				BeanDefinition customDefinition = parserContext.getDelegate()
						.parseCustomElement(eCustomBean,
								builder.getBeanDefinition());
				builder.addPropertyValue(propertyName, customDefinition);
			}
		}
		// Simple attributes
		for (Map.Entry<String, String> entry : this.simpleAttributes.entrySet()) {
			String attributeName = entry.getKey();
			String propertyName = entry.getValue();
			setAttributeAsProperty(element, attributeName, propertyName,
					builder);
		}
		// Fixed properties
		for (Map.Entry<String, String> entry : this.fixedProperties.entrySet()) {
			String propertyName = entry.getKey();
			String propertyValue = entry.getValue();
			setPropertyValue(element, propertyName, propertyValue, builder);
		}
		// Attributes refs
		for (Map.Entry<String, String> entry : this.attributeRefs.entrySet()) {
			String attributeName = entry.getKey();
			String propertyName = entry.getValue();
			setAttributeAsReference(element, attributeName, propertyName,
					builder);
		}
		// Custom beans
		List<Element> children = DOMUtils.getElements(element);
		for (Element eChild : children) {
			String ns = eChild.getNamespaceURI();
			String name = eChild.getLocalName();
			if (CustomNamespaces.isCustomNamespace(ns)) {
				/**
				 * Parses the custom bean if: <li>not defined as a custom list
				 * of strings <li>not defined as a custom bean
				 */
				if ((this.listStringContainer == null || !StringUtils.equals(
						this.listStringContainer.getFirstValue(), name))
						&& (this.listContainer == null || !StringUtils.equals(
								this.listContainer.getFirstValue(), name))
						&& (this.mapContainer == null || !StringUtils.equals(
								this.mapContainer.getFirstValue(), name))
						&& !this.customBeanAs.containsKey(name)) {
					parserContext.getDelegate().parseCustomElement(eChild,
							builder.getBeanDefinition());
				}
			}
		}
		// Attribute factories
		for (Map.Entry<String, List<Pair<String, SimpleBeanFactory>>> entry : this.attributeFactories
				.entrySet()) {
			String attributeName = entry.getKey();
			List<Pair<String, SimpleBeanFactory>> list = entry.getValue();
			for (Pair<String, SimpleBeanFactory> value : list) {
				String propertyName = value.getFirstValue();
				SimpleBeanFactory factory = value.getSecondValue();
				String attributeValue = element.getAttribute(attributeName);
				if (StringUtils.isNotBlank(attributeValue)) {
					Object bean = factory.create(attributeValue);
					if (bean != null) {
						builder.addPropertyValue(propertyName, bean);
					}
				}
			}
		}
		// List
		if (this.listContainer != null) {
			String childName = this.listContainer.getFirstValue();
			String propertyName = this.listContainer.getSecondValue();
			setListAsProperty(element, childName, propertyName, parserContext,
					builder);
		}
		// String list
		if (this.listStringContainer != null) {
			String childName = this.listStringContainer.getFirstValue();
			String propertyName = this.listStringContainer.getSecondValue();
			setListStringAsProperty(element, childName, propertyName,
					parserContext, builder);
		}
		// Map
		if (this.mapContainer != null) {
			String childName = this.mapContainer.getFirstValue();
			String propertyName = this.mapContainer.getSecondValue();
			setMapAsProperty(element, childName, propertyName, parserContext,
					builder);
		}
		// OK
		super.doParse(element, parserContext, builder);
		// Default ID
		if (StringUtils.isNotBlank(this.defaultID)) {
			String id = DOMUtils.getAttribute(element, "id", null);
			if (StringUtils.isBlank(id)) {
				element.setAttribute("id", this.defaultID);
			}
		}
		// Container property
		BeanDefinition containingBeanDefinition = parserContext
				.getContainingBeanDefinition();
		if (StringUtils.isNotBlank(this.containerProperty)
				&& containingBeanDefinition != null) {
			boolean contained = DOMUtils.getBooleanAttribute(element,
					"contained", false, true);
			if (contained) {
				containingBeanDefinition.getPropertyValues().addPropertyValue(
						this.containerProperty, builder.getBeanDefinition());
			}
		}
	}

	@Override
	protected Class<C> getBeanClass(Element element) {
		return this.beanClass;
	}

	/**
	 * Specifies that the parsed bean must be used as a property in the
	 * containing bean.
	 * 
	 * @param property
	 *            Name of property that must contain the parsed bean in the
	 *            containing bean
	 * @return This parser (allows chaining)
	 */
	public SimpleDefinitionParser<C> setContainerProperty(String property) {
		this.containerProperty = property;
		return this;
	}

	/**
	 * Sets the default ID to use when none is provided
	 * 
	 * @param defaultID
	 *            Default ID
	 * @return This instance
	 */
	public SimpleDefinitionParser<C> setDefaultID(String defaultID) {
		this.defaultID = defaultID;
		return this;
	}

	/**
	 * The child element defined by <code>childName</code> defines a list of
	 * objects which will be put in the <code>propertyName</code> property of
	 * the parsed bean.
	 * 
	 * @param childName
	 *            Name of the children that define the list
	 * @param propertyName
	 *            Property that must hold the list
	 * @return This parser (allows chaining)
	 */
	public SimpleDefinitionParser<C> setList(String childName,
			String propertyName) {
		this.listContainer = new Pair<String, String>(childName, propertyName);
		return this;
	}

	/**
	 * The child element defined by <code>childName</code> defines a list of
	 * {@link String}s which will be put in the <code>propertyName</code>
	 * property of the parsed bean.
	 * 
	 * @param childName
	 *            Name of the children that define the list of {@link String}s
	 * @param propertyName
	 *            Property that must hold the list
	 * @return This parser (allows chaining)
	 */
	public SimpleDefinitionParser<C> setListString(String childName,
			String propertyName) {
		this.listStringContainer = new Pair<String, String>(childName,
				propertyName);
		return this;
	}

	/**
	 * The child element defined by <code>childName</code> is a map (containing
	 * &lt;entry&gt;s elements) that will be stored in the
	 * <code>propertyName</code> property of the parsed bean.
	 * 
	 * @param childName
	 *            Name of the child that contains the map
	 * @param propertyName
	 *            Property that must hold the map
	 * @return This parser (allows chaining)
	 */
	public SimpleDefinitionParser<C> setMap(String childName,
			String propertyName) {
		this.mapContainer = new Pair<String, String>(childName, propertyName);
		return this;
	}

}
